@@ -1,5 +1,7 @@
 import math
+import copy
 import random
+import warnings
 from tqdm import tqdm
 import numpy as np
 from . import activation_function
@@ -53,10 +55,13 @@ class MLP:
         self.min_valid_loss = math.inf
         self.min_valid_cnt = 0
         self.validation_set_early_stop = False
+        self.hidden_weight_early_stop_memory = copy.deepcopy(self.hidden_weight)
+        self.output_weight_early_stop_memory = copy.deepcopy(self.output_weight)
+        self.early_stop_memory = False
 
 
 
-    def train(self, x_train, y_train, x_valid='NULL', y_valid='NULL', lr=0.01, epochs=100, batch_size=1, verbose_frequence=1, methods=False):
+    def train(self, x_train, y_train, x_valid='NULL', y_valid='NULL', lr=0.01, epochs=100, batch_size=1, verbose_frequency=1, methods=False):
         if len(x_train) % batch_size != 0:
             raise ValueError(f"batch_size需为训练集样本数量的约数")
 
@@ -65,20 +70,45 @@ class MLP:
                 if not (method == 'lr_decay' or method == 'validation_set_early_stop'):
                     raise ValueError(f"目前methods支持lr_decay及validation_set_early_stop")
                 if method == 'lr_decay':
+                    """
+                    if not methods['lr_decay'][0] == 'exp_decay' \
+                            and not methods['lr_decay'][0] == 'decay_based_on_validation_set':
+                        raise ValueError(f"目前lr_decay支持exp_decay与decay_based_on_validation_set")
+                    """
+
+                    if not isinstance(methods['lr_decay'], list):
+                        raise ValueError(f"lr_decay的值应为一个2维列表，形如['exp_decay', 0.01]")
+                    if len(methods['lr_decay']) != 2:
+                        raise ValueError(f"lr_decay的值应为一个2维列表，形如['exp_decay', 0.01]")
                     if not methods['lr_decay'][0] == 'exp_decay':
                         raise ValueError(f"目前lr_decay支持exp_decay")
                     else:
-                        if not isinstance(methods['lr_decay'][1], (int, float)):
-                            raise ValueError(f"列表第二列请输入正确的整形或字符型衰减率")
+                        if methods['lr_decay'][0] == 'exp_decay':
+                            if not isinstance(methods['lr_decay'][1], (int, float)):
+                                raise ValueError(f"列表第二列请输入正确的整形或字符型衰减率")
+                        """
+                        if methods['lr_decay'][0] == 'decay_based_on_validation_set':
+                            if not isinstance(methods['lr_decay'][1], int) \
+                                    or not isinstance(methods['lr_decay'][2], float):
+                                raise ValueError(f"列表第二列请输入正确的整形代数，第三列请输入正确的浮点型学习率衰减倍数")
+                            if methods['lr_decay'][2] <= 0 or methods['lr_decay'][2] >= 1:
+                                warnings.warn("衰减倍数可能不合法", UserWarning)
+                        """
+
                     lr0 = lr
                 if method == 'validation_set_early_stop':
+                    if not isinstance(methods['validation_set_early_stop'], list):
+                        raise ValueError(f"validation_set_early_stop的值应为一个2维列表，形如[50, True]")
+                    if len(methods['validation_set_early_stop']) != 2:
+                        raise ValueError(f"validation_set_early_stop的值应为一个2维列表，形如[50, True]")
                     if not isinstance(methods['validation_set_early_stop'][0], int):
                         raise ValueError(f"请输入合法的验证集提前终止代数（整形）")
                     else:
-                        if not methods['validation_set_early_stop'][1] is False:
-                            raise ValueError(f"目前只支持保存提前终止后训练的最后一代模型的参数，验证集损失最小的模型暂不支持")
+                        if not isinstance(methods['validation_set_early_stop'][1], bool):
+                            raise ValueError(f"validation_set_early_stop的值列表第二位请输入True或False以确定是否保存最优参数")
                         else:
                             self.validation_set_early_stop = methods['validation_set_early_stop'][0]
+                            self.early_stop_memory = methods['validation_set_early_stop'][1]
 
         ac = getattr(activation_function, self.hidden_activation)
         ac_derivative = getattr(activation_function, self.hidden_activation + '_derivative')
@@ -210,7 +240,7 @@ class MLP:
                         valid_epoch_loss += cross_entropy_loss  # 累加损失
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
                     self.valid_accuracy_list.append(valid_cnt / len(x_valid))
-                if (_ + 1) % verbose_frequence == 0:
+                if (_ + 1) % verbose_frequency == 0:
                     text += f'损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
                         text += f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%"
@@ -220,10 +250,16 @@ class MLP:
                     if self.valid_loss_values[-1] < self.min_valid_loss:
                         self.min_valid_cnt = 0
                         self.min_valid_loss = self.valid_loss_values[-1]
+                        if self.early_stop_memory is True:
+                            self.hidden_weight_early_stop_memory = copy.deepcopy(self.hidden_weight)
+                            self.output_weight_early_stop_memory = copy.deepcopy(self.output_weight)
                     else:
                         self.min_valid_cnt += 1
 
                 if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    if self.early_stop_memory is True:
+                        self.hidden_weight = copy.deepcopy(self.hidden_weight_early_stop_memory)
+                        self.output_weight = copy.deepcopy(self.output_weight_early_stop_memory)
                     print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
                     break
 
@@ -387,7 +423,7 @@ class MLP:
                         valid_epoch_loss += categorical_cross_entropy_loss  # 累加损失
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
                     self.valid_accuracy_list.append(valid_cnt / len(x_valid))
-                if (_ + 1) % verbose_frequence == 0:
+                if (_ + 1) % verbose_frequency == 0:
                     text += f'损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
                         text += f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%"
@@ -397,10 +433,16 @@ class MLP:
                     if self.valid_loss_values[-1] < self.min_valid_loss:
                         self.min_valid_cnt = 0
                         self.min_valid_loss = self.valid_loss_values[-1]
+                        if self.early_stop_memory is True:
+                            self.hidden_weight_early_stop_memory = copy.deepcopy(self.hidden_weight)
+                            self.output_weight_early_stop_memory = copy.deepcopy(self.output_weight)
                     else:
                         self.min_valid_cnt += 1
 
                 if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    if self.early_stop_memory is True:
+                        self.hidden_weight = copy.deepcopy(self.hidden_weight_early_stop_memory)
+                        self.output_weight = copy.deepcopy(self.output_weight_early_stop_memory)
                     print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
                     break
 
@@ -521,7 +563,7 @@ class MLP:
 
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
 
-                if (_ + 1) % verbose_frequence == 0:
+                if (_ + 1) % verbose_frequency == 0:
                     text += f'损失：{self.loss_values[-1]}'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
                         text += f"，验证集损失：{self.valid_loss_values[-1]}"
@@ -531,10 +573,16 @@ class MLP:
                     if self.valid_loss_values[-1] < self.min_valid_loss:
                         self.min_valid_cnt = 0
                         self.min_valid_loss = self.valid_loss_values[-1]
+                        if self.early_stop_memory is True:
+                            self.hidden_weight_early_stop_memory = copy.deepcopy(self.hidden_weight)
+                            self.output_weight_early_stop_memory = copy.deepcopy(self.output_weight)
                     else:
                         self.min_valid_cnt += 1
 
                 if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    if self.early_stop_memory is True:
+                        self.hidden_weight = copy.deepcopy(self.hidden_weight_early_stop_memory)
+                        self.output_weight = copy.deepcopy(self.output_weight_early_stop_memory)
                     print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
                     break
 
@@ -672,7 +720,7 @@ class MLP:
 
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
 
-                if (_ + 1) % verbose_frequence == 0:
+                if (_ + 1) % verbose_frequency == 0:
                     text += f'损失：{self.loss_values[-1]}'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
                         text += f"，验证集损失：{self.valid_loss_values[-1]}"
@@ -682,10 +730,16 @@ class MLP:
                     if self.valid_loss_values[-1] < self.min_valid_loss:
                         self.min_valid_cnt = 0
                         self.min_valid_loss = self.valid_loss_values[-1]
+                        if self.early_stop_memory is True:
+                            self.hidden_weight_early_stop_memory = copy.deepcopy(self.hidden_weight)
+                            self.output_weight_early_stop_memory = copy.deepcopy(self.output_weight)
                     else:
                         self.min_valid_cnt += 1
 
                 if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    if self.early_stop_memory is True:
+                        self.hidden_weight = copy.deepcopy(self.hidden_weight_early_stop_memory)
+                        self.output_weight = copy.deepcopy(self.output_weight_early_stop_memory)
                     print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
                     break
 
@@ -923,7 +977,7 @@ class SOM:
         self.SOM_weight = SOM_weight
         self.dist_values = []
 
-    def train(self, x, lr=0.01, epochs=100, start_range=4, verbose_frequence=1):
+    def train(self, x, lr=0.01, epochs=100, start_range=4, verbose_frequency=1):
 
         for _ in tqdm(range(epochs)):
             field_range = basic_function.decay(_, epochs, start_range)
@@ -948,7 +1002,7 @@ class SOM:
                         for dim in range(self.input_node_num):
                             self.SOM_weight[j][k][dim] += lr * basic_function.hci(node_dist, field_range) * (
                                         x[i][dim] - self.SOM_weight[j][k][dim])
-            if (_ + 1) % verbose_frequence == 0:
+            if (_ + 1) % verbose_frequency == 0:
                 print(f"第{_ + 1}代平均最小距离：{sum(min_dist_list) / len(x)}")
             self.dist_values.append(sum(min_dist_list) / len(x))
         return self.dist_values
